@@ -139,13 +139,11 @@ typedef int ssize_t;
 #define PATH_TMP        "/tmp"
 #define PATH_VAR_TMP    "/var/tmp"
 #define PATH_PASSWD     "/etc/passwd"
-#define PATH_WTMP       "/var/adm/wtmp"
-#define PATH_UTMP       "/var/adm/utmp"
+#define PATH_WTMP       "/var/adm/wtmpx"
+#define PATH_UTMP       "/var/adm/utmpx"
 #define PATH_SYSLOG     "/var/adm/messages"
-#ifdef S_ISSOCK
-#undef S_ISSOCK
-#define       S_ISSOCK        S_ISFIFO
-#endif
+#define IS_UNIX_SOCKET(m)  (S_ISSOCK(m) || S_ISFIFO(m))
+#include <sys/resource.h>
 int getrusage(int who, struct rusage *rusage);
 #endif
 
@@ -241,6 +239,24 @@ int getrusage(int who, struct rusage *rusage);
 #define PATH_MAILLOG	"/var/spool/mqueue/syslog"
 #endif
 
+#if defined(_AIX41) || defined (AIX41)
+#define HAVE_GETRUSAGE
+#if defined(__64BIT_KERNEL) && defined(_KERNEL)
+#define PID_T_IS_LONG_INT
+#endif /* __64BIT_KERNEL && _KERNEL */
+#define NEED_SYS_SELECT_H
+#define SIZEOF_IS_LONG_INT
+#define SOCKLEN_T	int
+#define O_RSYNC		0
+#define PATH_TMP	"/tmp"
+#define PATH_VAR_TMP	"/var/tmp"
+#define PATH_PASSWD	"/etc/passwd"
+#define PATH_WTMP	"/var/adm/wtmp"
+#define PATH_UTMP	"/etc/utmp"
+#define PATH_SYSLOG	"/var/adm/messages"
+#define PATH_MAILLOG	"/var/spool/mqueue/syslog"
+#endif /* _AIX41 || AIX41 */
+
 #ifdef UNIXWARE7
 #define HAVE_GETRUSAGE
 #define HAVE_SNPRINTF
@@ -258,10 +274,7 @@ int getrusage(int who, struct rusage *rusage);
  * Strange thing on Unixware 7, sockets manifest itself as FIFOs
  * John Hughes <john@Calva.COM>
  */
-#ifdef S_ISSOCK
-#undef S_ISSOCK
-#define	S_ISSOCK	S_ISFIFO
-#endif
+#define IS_UNIX_SOCKET(m)  (S_ISSOCK(m) || S_ISFIFO(m))
 #endif
 
 #ifdef OpenUNIX8
@@ -277,10 +290,7 @@ int getrusage(int who, struct rusage *rusage);
 #define PATH_WTMP       "/var/adm/wtmp"
 #define PATH_UTMP       "/etc/utmp"
 #define PATH_SYSLOG     "/var/adm/syslog"
-#ifdef S_ISSOCK
-#undef S_ISSOCK
-#define	S_ISSOCK	S_ISFIFO
-#endif
+#define IS_UNIX_SOCKET(m)  (S_ISSOCK(m) || S_ISFIFO(m))
 #endif
 
 #ifdef SCO
@@ -304,10 +314,7 @@ extern void closelog(void);
 #define PATH_WTMP       "/var/adm/wtmp"
 #define PATH_UTMP       "/etc/utmp"
 #define PATH_SYSLOG     "/var/adm/syslog"
-#ifdef S_ISSOCK
-#undef S_ISSOCK
-#define	S_ISSOCK	S_ISFIFO
-#endif
+#define IS_UNIX_SOCKET(m)  (S_ISSOCK(m) || S_ISFIFO(m))
 #endif
 
 #ifdef SUNOS4
@@ -380,10 +387,7 @@ extern closelog(void);
 #define PATH_WTMP       "/var/adm/wtmpx"
 #define PATH_UTMP       "/var/adm/utmpx"
 #define PATH_SYSLOG     "/var/adm/log/messages"
-#ifdef S_ISSOCK
-#undef S_ISSOCK
-#define       S_ISSOCK        S_ISFIFO
-#endif
+#define IS_UNIX_SOCKET(m)  (S_ISSOCK(m) || S_ISFIFO(m))
 #endif
 
 #ifndef RANDSAVENAME
@@ -402,8 +406,16 @@ extern closelog(void);
 /*
  * If not available, we cannot perform the check and will always succeed.
  */
-#ifndef S_ISSOCK
-#define S_ISSOCK(m)	(1)
+#ifndef IS_UNIX_SOCKET
+# ifdef S_ISSOCK
+#  define IS_UNIX_SOCKET(m)	(S_ISSOCK(m))
+# else
+#  define IS_UNIX_SOCKET(m)	(1)
+# endif
+#else
+# ifndef S_ISSOCK
+#  define S_ISSOCK		(1)
+# endif
 #endif
 
 /*
