@@ -369,7 +369,7 @@ int main(int argc, char *argv[])
         killfail = 1;
         continue;
       }
-      if (!S_ISSOCK(randsock_file_stat.st_mode)) {
+      if (!IS_UNIX_SOCKET(randsock_file_stat.st_mode)) {
 	(void)snprintf(msg_buf, sizeof(msg_buf), "Will not touch %s: no socket",
 		randsock_file[i]);
 	msg_buf[sizeof(msg_buf) - 1] = '\0';
@@ -440,7 +440,7 @@ int main(int argc, char *argv[])
       exit(EXIT_FAILURE);
     }
     else if (!statval) {
-      if (!S_ISSOCK(randsock_file_stat.st_mode)) {
+      if (!IS_UNIX_SOCKET(randsock_file_stat.st_mode)) {
 	(void)snprintf(msg_buf, sizeof(msg_buf), "Will not touch %s: no socket",
 		randsock_file[i]);
 	msg_buf[sizeof(msg_buf) - 1] = '\0';
@@ -471,6 +471,13 @@ int main(int argc, char *argv[])
     }
 
     service_socket[service_count++] = unix_listen(randsock_file[i], 1, 1);
+    if (chmod(randsock_file[i], socketmode) < 0) {
+      (void)snprintf(msg_buf, sizeof(msg_buf), "Could not chmod socket to %04o",
+		     socketmode);
+      msg_buf[sizeof(msg_buf) - 1] = '\0';
+      error_fatal(msg_buf, errno);
+      exit(EXIT_FAILURE);
+    }
     count_openfd++;
   }
 
