@@ -15,9 +15,10 @@
 
 #include "prngd.h"
 
-static volatile seed_t gatherer_seed;
-static volatile pid_t sigchild_pid = -1;
-static volatile int gatherer_status;
+#ifdef NEED_SYS_SELECT_H
+#include <sys/select.h>
+#endif /* NEED_SYS_SELECT_H */
+
 static volatile int caught_sigchild = 0;
 static volatile int received_sighup = 0;
 
@@ -55,32 +56,45 @@ static void sighup_handler(int received_signal /*ARGSUSED*/)
 
 static void child_handler(int received_signal /*ARGSUSED*/)
 {
-  int save_errno = errno;
-  int my_gatherer_status;
-  pid_t pid;
-
   /*
    * Wait for child process to avoid zombies. This signal comes in
-   * asynchronously and only few functions are safe to be called,
-   * especially stdio and syslog functions cannot be called.
+   * asynchronously.
    *
    * PRNGD can only have one child at a time, so recording the information
-   * is trivial.
+   * could be trivial. However: if a gatherer call is stuck inside a system
+   * call, the processes will only die much later, even though "kill -9"
+   * has been called after PRNGD's gatherer timeout. Thus, several children
+   * may die at the same time. On the other hand, due to process scheduling,
+   * a gatherer may die before its PID is recorded. As we must not miss the
+   * exit of the actual gatherer process (so that it can be killed after
+   * timeout), we only catch SIGCHLD but the evaluation is done by a
+   * seperate polling routine that does know about the active gatherer.
    */
   caught_sigchild = 1;
-
-  do {
-    pid = waitpid(-1, &my_gatherer_status, WNOHANG);
-  } while (pid == -1 && errno == EINTR);
-
-  if (WIFEXITED(my_gatherer_status) || WIFSIGNALED(my_gatherer_status)) {
-    sigchild_pid = pid;
-    gatherer_status = my_gatherer_status;
-  }
-
-  errno = save_errno;
 }
 
+static void collect_children(int *gatherer_pid)
+{
+  int child_pid;
+  int gatherer_status;
+
+  while (((child_pid = waitpid(-1, &gatherer_status, WNOHANG)) > 0)
+	 || (errno == EINTR)) {
+    if (child_pid > 0) {
+      rand_add((int *)&gatherer_status, sizeof(gatherer_status), 0.0);
+      if (WIFEXITED(gatherer_status) || WIFSIGNALED(gatherer_status)) {
+        if (debug)
+	  (void)fprintf(stderr,
+		      "Found child died (pid = %ld, gatherer_pid = %ld)\n",
+		      (long)child_pid, (long)*gatherer_pid);
+        if (child_pid == *gatherer_pid)
+	*gatherer_pid = -1;
+      }
+    }
+  }
+}
+
+
 void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
 	       time_t seed_ext_interval, entropy_source_t *entropy_source,
 	       int max_gatherer_bytes)
@@ -97,11 +111,13 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
   egd_query_t *egd_query = NULL;
   egd_query_t *egd_query_old;
   int max_fd;
-  pid_t gatherer_pid, child_pid;
+  int max_query_old;
+  pid_t gatherer_pid;
   int gather_fd;
   int gatherer_entry;
   int gatherer_bytes;
   int first_gather = 1;
+  int retrieved_bytes = 0;
   double est = 0.0;
   time_t gather_start;
   unsigned char gatherbuf[10240];
@@ -196,7 +212,8 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
      * If we just entered the loop and gather for the first time, we
      * don't want to wait.
      */
-    if (first_gather || (rand_pool() < (ENTROPY_THRESHOLD))) {
+    if (first_gather || (rand_pool() < (ENTROPY_THRESHOLD)) || retrieved_bytes)
+    {
       timeout.tv_sec = 0;
       timeout.tv_usec = 10;
     }
@@ -204,27 +221,14 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       timeout.tv_sec = (seed_stat_interval < seed_ext_interval ?
 			 seed_stat_interval : seed_ext_interval);
       timeout.tv_usec = 0;
+      if ((gatherer_pid > 0) && (gather_fd < 0))
+	timeout.tv_sec = 1;	/* if we are still waiting for a child, don't
+				   sit long in select() */
     }
     num = select(max_fd + 1, &rfds, &wfds, &efds, &timeout);
     saved_errno = errno;
 
-    /*
-     * Depending on the scheduling the SIGCHILD may already be old. This
-     * does not matter however, since we only have one child at a time
-     * and do not start another one before the old one has been killed
-     * (or at least tried to with SIGKILL).
-     */
-    if (caught_sigchild) {
-      caught_sigchild = 0;
-      child_pid = sigchild_pid;		/* shut up warnings, as sigchild_pid */
-					/* is volatile */
-      if (debug)
-	(void)fprintf(stderr,
-		      "Caught SIGCHLD (pid = %ld, gatherer_pid = %ld)\n",
-		      (long)child_pid, (long)gatherer_pid);
-      if (child_pid == gatherer_pid)
-	gatherer_pid = -1;
-    }
+    collect_children(&gatherer_pid);
 
     if (received_sighup) {
       received_sighup = 0;
@@ -248,6 +252,81 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       }
     }
 
+    /*
+     * First thing to do: check, whether new connections are waiting.
+     * The queue on the listen() socket is of limited length, so
+     * querying processes might run into "connection refused" errors,
+     * when the queuelength is exceeded.
+     */
+    max_query_old = max_query;
+    for (i = 0; i < numsock; i++)
+    {
+      if (count_openfd >= limit_openfd - OPEN_FD_RESERVE) {
+	(void)snprintf(msg_buf, sizeof(msg_buf),
+		"could not accept connection: no more slots (%d of %d)\n",
+		count_openfd, limit_openfd - OPEN_FD_RESERVE);
+	msg_buf[sizeof(msg_buf) - 1] = '\0';
+	error_notice(msg_buf, 0);
+	continue;	/* do not accept new connection */
+      }
+      if (FD_ISSET(service_socket[i], &rfds))
+      {
+        addrlen = (SOCKLEN_T) sizeof(addr);
+        accept_fd = accept(service_socket[i], &addr, &addrlen);
+        if (accept_fd < 0)
+	  continue;		/*
+				 * Ignore and stay alive 
+				 */
+
+	count_openfd++;
+	non_blocking(accept_fd, 1);
+	close_on_exec(accept_fd, 1);
+
+        /*
+         * Create a new structure for the incoming request 
+         */
+        if (max_query + 1 > alloc_query)
+        {
+	  egd_query_old = egd_query;
+	  if (alloc_query == 0)
+	    egd_query = malloc((max_query + 1) * sizeof(egd_query_t));
+	  else
+	    egd_query =
+	        realloc(egd_query_old,
+		        (max_query + 1) * sizeof(egd_query_t));
+	  if (!egd_query)
+	  {
+	    error_error("realloc() failed", errno);
+	    (void)close(accept_fd);	/*
+				 * Better this than complete shut down 
+				 */
+	    egd_query = egd_query_old;
+	    continue;
+	  }
+	  alloc_query = max_query + 1;
+        }
+        /*
+         * This routine is safe here, as in case of error we do "continue;"
+         * in case of problems. So we have succesfully reached this point.
+         */
+        max_query++;
+        egd_query[max_query - 1].fd = accept_fd;
+        egd_query[max_query - 1].state = wait_command;
+	if (debug) {
+          (void)fprintf(stderr, "Opening service fd %d\n",
+	  	egd_query[max_query-1].fd);
+	}
+	/*
+	 * Finally, mix information about the connection into the pool.
+	 * Maybe its coming from TCP and has some foreign port number
+	 * in it. Only count it as '0', because maybe there is no
+	 * entropy in it at all.
+	 */
+	rand_add(&addr, addrlen, 0.0);
+      }
+    }
+
+
     if (gather_fd != -1) {
       if (FD_ISSET(gather_fd, &rfds)) {
 	/*
@@ -263,8 +342,6 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
 	  if (debug)
 	    (void)fprintf(stderr, "Closing connection to gatherer\n");
 	  (void)close(gather_fd);
-	  rand_add((int *)&gatherer_status, sizeof(gatherer_status), 0.0);
-	  rand_add((seed_t *)&gatherer_seed, sizeof(seed_t), 0.25);
 	  gather_fd = -1;
 	  count_openfd--;
 	}
@@ -289,24 +366,8 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       }
     }
 
-    /*
-     * If a gatherer has not finished in time, kill it!! Re-check, whether
-     * the signal came in in between. This might happen, since we close
-     * the connection when too many bytes where sent. If SIGCHILD is not
-     * there and the time is over, we kill anyway. If the time is not over,
-     * the check will be done at the next possible occation.
-     */
-    if (caught_sigchild) {
-      caught_sigchild = 0;
-      child_pid = sigchild_pid;		/* shut up warnings, as sigchild_pid */
-					/* is volatile */
-      if (debug)
-	(void)fprintf(stderr,
-		      "Caught SIGCHLD (pid = %ld, gatherer_pid = %ld)\n",
-		      (long)child_pid, (long)gatherer_pid);
-      if (child_pid == gatherer_pid)
-	gatherer_pid = -1;
-    }
+    collect_children(&gatherer_pid);
+
     if ((gatherer_pid > 0) &&
 	(difftime(time(NULL), gather_start) > seed_ext_interval / 2 + 1)) {
       if (debug)
@@ -328,6 +389,8 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       }
     }
 
+    collect_children(&gatherer_pid);
+
     /*
      * Call a gatherer, if it is time to do so. When starting up, call
      * all gatherers once to obtain initial seeding.
@@ -417,16 +480,28 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
 	  }
 	  (void)fprintf(stderr, ")\n");
 	}
+	collect_children(&gatherer_pid);
       }
       continue;
 
     }
     /*
-     * Do service the active queries first, because accepting a new query
-     * will create a new fd. First step: check errors reported and mark
-     * the corresponding descriptor "to_be_error_closed".
+     * Finally, serve the active sockets. First step: check errors reported
+     * and mark the corresponding descriptor "to_be_error_closed". Perform
+     * the loop only over the sockets already open, as the data returned
+     * from select() obviously cannot make sense for the connections
+     * freshly opened.
+     * Service is in the sequence the fds were opened, but there is one
+     * exception: during the generation of the random numbers inside the
+     * PRNG, no other operation can be performed. It is therefore not
+     * good to generate too much random numbers inside the loop without
+     * actually leaving the loop. Incoming connections may pend on the
+     * listen socket. We will thus leave the loop as soon as we retrieved
+     * random numbers from the PNRG and run through the complete serverloop
+     * once again.
      */
-    for (i = 0; i < max_query; i++)
+    for (i = retrieved_bytes = 0;
+	 (i < max_query_old) && (!retrieved_bytes); i++)
     {
       if (FD_ISSET(egd_query[i].fd, &efds))
       {
@@ -443,8 +518,10 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       case wait_entropy:
 	  seed_internal(NULL);	/* Mix in just before query */
 	  seed_stat();
-	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
+	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0) {
 	    egd_query[i].state = wait_write;
+	    retrieved_bytes = 1;
+	  }
 	break;
       case wait_command:
 	error = socket_read(egd_query[i].fd, &(egd_query[i].command), 1);
@@ -486,8 +563,10 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
 	{
 	  seed_internal(NULL);	/* Mix in just before query */
 	  seed_stat();
-	  if (rand_bytes(egd_query[i].buffer + 1, egd_query[i].secbyte))
+	  if (rand_bytes(egd_query[i].buffer + 1, egd_query[i].secbyte)) {
 	    numbytes = egd_query[i].secbyte;
+	    retrieved_bytes = 1;
+	  }
 	  else
 	    numbytes = 0;
 	  egd_query[i].buffer[0] = (char)numbytes;
@@ -501,8 +580,10 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
 	  egd_query[i].yet_processed = 0;
 	  seed_internal(NULL);	/* Mix in just before query */
 	  seed_stat();
-	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
+	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0) {
 	    egd_query[i].state = wait_write;
+	    retrieved_bytes = 1;
+	  }
 	  else
 	    egd_query[i].state = wait_entropy;
 	}
@@ -568,6 +649,9 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       }
       }
     }
+    /*
+     * Perform the final cleanup
+     */
     for (i = 0; i < max_query; i++) {
       if ((egd_query[i].state == to_be_closed) ||
 	  (egd_query[i].state == to_be_error_closed)) {
@@ -595,72 +679,8 @@ void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
       }
     }
 
-    for (i = 0; i < numsock; i++)
-    {
-      if (count_openfd >= limit_openfd - OPEN_FD_RESERVE) {
-	(void)snprintf(msg_buf, sizeof(msg_buf),
-		"could not accept connection: no more slots (%d of %d)\n",
-		count_openfd, limit_openfd - OPEN_FD_RESERVE);
-	msg_buf[sizeof(msg_buf) - 1] = '\0';
-	error_notice(msg_buf, 0);
-	continue;	/* do not accept new connection */
-      }
-      if (FD_ISSET(service_socket[i], &rfds))
-      {
-        addrlen = sizeof(addr);
-        accept_fd = accept(service_socket[i], &addr, &addrlen);
-        if (accept_fd < 0)
-	  continue;		/*
-				 * Ignore and stay alive 
-				 */
 
-	count_openfd++;
-	non_blocking(accept_fd, 1);
-	close_on_exec(accept_fd, 1);
-
-        /*
-         * Create a new structure for the incoming request 
-         */
-        if (max_query + 1 > alloc_query)
-        {
-	  egd_query_old = egd_query;
-	  if (alloc_query == 0)
-	    egd_query = malloc((max_query + 1) * sizeof(egd_query_t));
-	  else
-	    egd_query =
-	        realloc(egd_query_old,
-		        (max_query + 1) * sizeof(egd_query_t));
-	  if (!egd_query)
-	  {
-	    error_error("realloc() failed", errno);
-	    (void)close(accept_fd);	/*
-				 * Better this than complete shut down 
-				 */
-	    egd_query = egd_query_old;
-	    continue;
-	  }
-	  alloc_query = max_query + 1;
-        }
-        /*
-         * This routine is safe here, as in case of error we do "continue;"
-         * in case of problems. So we have succesfully reached this point.
-         */
-        max_query++;
-        egd_query[max_query - 1].fd = accept_fd;
-        egd_query[max_query - 1].state = wait_command;
-	if (debug) {
-          (void)fprintf(stderr, "Opening service fd %d\n",
-	  	egd_query[max_query-1].fd);
-	}
-	/*
-	 * Finally, mix information about the connection into the pool.
-	 * Maybe its coming from TCP and has some foreign port number
-	 * in it. Only count it as '0', because maybe there is no
-	 * entropy in it at all.
-	 */
-	rand_add(&addr, addrlen, 0.0);
-      }
-    }
+    collect_children(&gatherer_pid);
 
   }
 
